# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, TypeAlias, final

import numpy as np
from loguru import logger

from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.pydantic_models.configs.observer_config import ObserverConfig
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import IncludeStatisticsType, TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates
from libinephany.utils.exceptions import InvalidObservationSizeError

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

CachedObservationData: TypeAlias = int | float | list[int | float] | TensorStatistics | dict[str, float | int]

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class Observer(ABC):

    TRAIN = "train"
    INFER = "infer"

    def __init__(
        self,
        *,
        observer_config: ObserverConfig,
        include_statistics: IncludeStatisticsType = None,
        include_hparams: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        :param global_config: ObserverConfig that can be used to inform various observation calculations.
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations, optionally with transform names.
        :param include_hparams: If the observation uses the HyperparameterStates model to return observations, names of the
        hyperparameters to include in returned observations.
        :param kwargs: Miscellaneous keyword arguments.
        """

        self._mode = self.TRAIN
        self._validated_observation = False
        self._cached_observation: CachedObservationData | None = None

        self.observer_config = observer_config
        self.invalid_observation_threshold = observer_config.invalid_observation_threshold

        self.include_statistics: IncludeStatisticsType = None
        self.include_hparams = include_hparams

        if include_statistics is not None:
            self.include_statistics = include_statistics

        if self.requires_include_statistics and not self.include_statistics:
            raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")

        if self.requires_include_hparams and not self.include_hparams:
            raise ValueError(f"{self.__class__.__name__} must be provided with include_hparams.")

    @final
    @property
    def in_training_mode(self) -> bool:
        """
        :return: Whether the observer is in training mode.
        """

        return self._mode == self.TRAIN

    @final
    @property
    def observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        observation_format = self._get_observation_format()

        if observation_format in StatisticStorageTypes:
            return observation_format

        raise ValueError(
            f"The type of observation format that observations should return must be in the "
            f"{StatisticStorageTypes.__name__} enumeration class. {observation_format} is not and so "
            f"{self.__class__.__name__} is invalid!"
        )

    @property
    def observation_size(self) -> int:
        """
        :return: Number of elements in the observation vector this observation takes up.
        """

        observation_format = self.observation_format

        if observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
            if self.include_statistics is None:
                raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")

            return len([field for field in TensorStatistics.model_fields.keys() if field in self.include_statistics])

        elif observation_format is StatisticStorageTypes.FLOAT:
            return 1

        elif observation_format is StatisticStorageTypes.VECTOR:
            return self.vector_length

        else:
            raise ValueError(f"{observation_format} is not a recognised format that observations can return!")

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """

        raise NotImplementedError

    @property
    def can_inform(self) -> bool:
        """
        :return: Whether observations from the observer can be used in the agent info dictionary.
        """

        return True

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return False

    @property
    def requires_include_hparams(self) -> bool:
        """
        :return: Whether the observation requires include_hparams to be provided.
        """

        return False

    @property
    def requires_validation_loss(self) -> bool:
        """
        :return: Whether the observation requires validation loss to be calculated.
        """

        return False

    @abstractmethod
    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        raise NotImplementedError

    @abstractmethod
    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalise loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        raise NotImplementedError

    @final
    def _validate_observation(self, observations: list[float | int]) -> None:
        """
        :param observations: Observation vector to validate.
        """

        if len(observations) != self.observation_size:
            raise InvalidObservationSizeError(
                f"The observation vector returned by {self.__class__.__name__} has length {len(observations)} but the "
                f"expected length is {self.observation_size}!"
            )

        self._validated_observation = True

    @final
    def _validate_observation_values(self, observations: list[float | int]) -> None:
        """
        :param observations: Observation vector to validate.
        """
        validate = True
        if len(observations) == 0:
            raise ValueError(f"Observer: {self.__class__.__name__} gathered observations with empty list!")

        if any(not isinstance(observation, (float, int)) for observation in observations):
            # Just in case
            other_count = sum(1 for observation in observations if not isinstance(observation, (float, int)))
            other_ratio = other_count / len(observations)
            raise ValueError(
                f"Observer: {self.__class__.__name__} gathered observations with invalid values (not float or int)! "
                f"Ratio: {other_ratio:.2%} ({other_count}/{len(observations)})"
            )

        # Check for NaN values
        if any(np.isnan(observation) for observation in observations):
            nan_count = sum(1 for observation in observations if np.isnan(observation))
            nan_ratio = nan_count / len(observations)
            logger.warning(
                f"Observer: {self.__class__.__name__} gathered observations with NaN values! Ratio: {nan_ratio:.2%} "
                f" ({nan_count}/{len(observations)})"
            )
            validate = False

        # Check for very large values (including negative values)
        if any(abs(observation) > self.invalid_observation_threshold for observation in observations):
            inf_count = sum(1 for observation in observations if observation > self.invalid_observation_threshold)
            inf_ratio = inf_count / len(observations)
            logger.warning(
                f"Observer: {self.__class__.__name__} gathered observations with large values (threshold: "
                f"{self.invalid_observation_threshold})! Ratio: {inf_ratio:.2%} ({inf_count}/{len(observations)})"
            )
            validate = False

        logger.trace(
            f'Observer: {self.__class__.__name__} observation validation {"passed" if validate else "failed"}!'
        )

    @abstractmethod
    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        raise NotImplementedError

    @final
    def observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        return_dict: bool = False,
        num_categories: int | None = None,
    ) -> tuple[list[float | int], dict[str, float] | None]:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param return_dict: Whether to return a dictionary of observations as well as the normal vector.
        :param num_categories: Number of categories, used to normalise loss based observations.
        :return: Tuple of:
            - List of floats or integers to add to the agent's observation vector.
            - Dictionary of specific observation values if the storage type is TensorStatistics and None otherwise.
        """

        observations_dict: dict[str, float] | None = None

        observations = self._observe(
            observation_inputs=observation_inputs,
            hyperparameter_states=hyperparameter_states,
            tracked_statistics=tracked_statistics,
            action_taken=action_taken,
            num_categories=num_categories,
        )

        if self.can_inform and self._cached_observation is None:
            self._cached_observation = deepcopy(observations)

        if self.observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
            if self.include_statistics is None:
                raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")

            if return_dict:
                observations_dict = observations.model_dump()  # type: ignore

            observations = observations.to_list(include_statistics=self.include_statistics)  # type: ignore

        observations = [observations] if not isinstance(observations, list) else observations  # type: ignore

        if not self._validated_observation:
            self._validate_observation(observations=observations)

        self._validate_observation_values(observations=observations)

        return observations, observations_dict

    @final
    def inform(self) -> float | int | dict[str, float] | None:
        """
        :return: The cached observation. If the observation format is TensorStatistics then it is converted to a
        dictionary with the statistics specified in include_statistics included.
        """

        if not self.can_inform:
            return None

        if self._cached_observation is None:
            raise ValueError(
                f"{self.__class__.__name__} cannot inform when no observation has been cached! Ensure observe(...) has "
                f"been called before inform(...)!"
            )

        if self.observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
            if self.include_statistics is None:
                raise ValueError(f"{self.__class__.__name__} must be provided with include_statistics.")

            observation = self._cached_observation.model_dump(include=set(self.include_statistics))  # type: ignore

        else:
            observation = self._cached_observation

        self._cached_observation = None

        return observation

    @final
    def train(self) -> None:
        """
        Sets the observer to train mode.
        """

        self._mode = self.TRAIN

    @final
    def infer(self) -> None:
        """
        Sets the observer to inference mode.
        """

        self._mode = self.INFER

    def reset(self) -> None:
        """
        Resets the observer.
        """


class LocalObserver(Observer, ABC):

    def __init__(
        self,
        *,
        agent_id: str,
        parameter_group_name: str | None,
        number_of_discrete_actions: int | None,
        action_scheme_index: int,
        **kwargs,
    ) -> None:
        """
        :param agent_id: ID of the agent this observation is assigned to.
        :param parameter_group_name: Name of the parameter group this agent is modulating.
        :param number_of_discrete_actions: The number of discrete actions that can be taken by this agent type's
        policy.
        :param action_scheme_index: Index of the action scheme in use in the list of possible action schemes.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(**kwargs)

        self.agent_id = agent_id
        self.parameter_group_name = parameter_group_name

        self.number_of_discrete_actions = number_of_discrete_actions
        self.action_scheme_index = action_scheme_index

    @abstractmethod
    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        raise NotImplementedError

    @abstractmethod
    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        raise NotImplementedError

    @abstractmethod
    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        raise NotImplementedError


class GlobalObserver(Observer, ABC):

    @abstractmethod
    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        raise NotImplementedError

    @abstractmethod
    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        raise NotImplementedError

    @abstractmethod
    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        raise NotImplementedError



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import numpy as np
from pydantic import BaseModel, ConfigDict

from libinephany.utils.typing import ObservationInformation

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class AgentInfo(BaseModel):

    model_config = ConfigDict(extra="allow")

    global_information: ObservationInformation = {}
    agent_information: dict[str, ObservationInformation] = {}
    global_critic_only_information: ObservationInformation = {}
    critic_only_agent_information: dict[str, ObservationInformation] = {}

    def _add_agent_information(self, agent_id: str, agent_information: ObservationInformation) -> None:
        """
        :param agent_id: ID of the agent that the given information is from.
        :param agent_information: Information from a particular agent to store.
        """

        if agent_id in self.agent_information:
            raise ValueError(f"Agent {agent_id} already exists within AgentInfo model!")

        self.agent_information[agent_id] = agent_information

    def add_agent_information(self, agent_information: dict[str, ObservationInformation]) -> None:
        """
        :param agent_information: Information from a particular agent to store.
        """

        for agent_id, single_agent_info in agent_information.items():
            self._add_agent_information(agent_id=agent_id, agent_information=single_agent_info)

    def _add_critic_only_agent_information(
        self, agent_id: str, critic_only_agent_information: ObservationInformation
    ) -> None:
        """
        :param agent_id: ID of the agent that the given information is from.
        :param critic_only_agent_information: Information from a particular agent to store.
        """

        if agent_id in self.critic_only_agent_information:
            raise ValueError(f"Agent {agent_id} already exists within AgentInfo model!")

        self.critic_only_agent_information[agent_id] = critic_only_agent_information

    def add_critic_only_agent_information(
        self, critic_only_agent_information: dict[str, ObservationInformation]
    ) -> None:
        """
        :param critic_only_agent_information: Information from a particular agent to store.
        """

        for agent_id, single_agent_info in critic_only_agent_information.items():
            self._add_critic_only_agent_information(agent_id=agent_id, critic_only_agent_information=single_agent_info)

    def update_global_information(self, new_information: ObservationInformation) -> None:
        """
        :param new_information: Additional global information to store.
        """

        for key, value in new_information.items():
            is_nan = False

            if not isinstance(value, dict):
                is_nan = np.isnan(value)

            if key in self.global_information and not isinstance(self.global_information[key], dict):
                is_nan = is_nan or np.isnan(self.global_information[key])  # type: ignore

            if key in self.global_information and not is_nan and value != self.global_information[key]:
                raise ValueError(
                    f"Value mismatch in global observations! Key {key} came with value {value} which conflicts with "
                    f"pre-existing value {self.global_information[key]}."
                )

            self.global_information[key] = value

    def update_global_critic_only_information(self, new_information: ObservationInformation) -> None:
        """
        :param new_information: Additional global information to store.
        """

        for key, value in new_information.items():
            is_nan = False

            if not isinstance(value, dict):
                is_nan = np.isnan(value)

            if key in self.global_critic_only_information and not isinstance(
                self.global_critic_only_information[key], dict
            ):
                is_nan = is_nan or np.isnan(self.global_critic_only_information[key])  # type: ignore

            if (
                key in self.global_critic_only_information
                and not is_nan
                and value != self.global_critic_only_information[key]
            ):
                raise ValueError(
                    f"Value mismatch in global observations! Key {key} came with value {value} which conflicts with "
                    f"pre-existing value {self.global_critic_only_information[key]}."
                )

            self.global_critic_only_information[key] = value



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from typing import Any

from libinephany.observations import observation_utils, statistic_trackers
from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class GlobalFirstOrderGradients(GlobalObserver):

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.FirstOrderGradients.__name__]

        return observation_utils.average_tensor_statistics(tensor_statistics=list(statistics.values()))  # type: ignore

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {statistic_trackers.FirstOrderGradients.__name__: dict(include_statistics=self.include_statistics)}


class GlobalSecondOrderGradients(GlobalObserver):
    # todo (tristan): we have reason to believe that this observer/tracker is wrong!
    # In fact has never been used correctly.
    # For now advised against using it

    def __init__(
        self,
        *,
        compute_hessian_diagonal: bool = False,
        **kwargs,
    ) -> None:
        """
        :param compute_hessian_diagonal: Whether to compute the Hessian diagonal to determine second order gradients
        or use the squared first order gradients as approximations in the same way Adam does.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(**kwargs)

        self.compute_hessian_diagonal = compute_hessian_diagonal

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.SecondOrderGradients.__name__]

        return observation_utils.average_tensor_statistics(tensor_statistics=list(statistics.values()))  # type: ignore

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {
            statistic_trackers.SecondOrderGradients.__name__: dict(
                include_statistics=self.include_statistics, compute_hessian_diagonal=self.compute_hessian_diagonal
            )
        }



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from typing import Any, Callable

from pydantic import BaseModel

from libinephany.utils.constants import POLICIES_TO_PREFIXES

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

VRAM_USAGES_KEY = "vram_usages"

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class InnerTaskProfile(BaseModel):

    inner_task_name: str

    number_of_agents: int
    number_of_layers: int
    number_of_parameters: int

    observation_space_sizes: dict[str, int]
    action_space_sizes: dict[str, int]

    vram_usages: dict[int, tuple[float, float]] | None = None
    times: dict[int, float] | None = None
    expected_vram_usage: float | None = None
    expected_idle_vram_usage: float | None = None
    max_batch_size_override: int | None = None

    @property
    def vram_usage(self) -> float:
        """
        :return: VRAM usage at the max batch size.
        """

        if self.expected_vram_usage is None:
            return float("nan")

        return self.expected_vram_usage

    @property
    def idle_vram_usage(self) -> float:
        """
        :return: Idle VRAM usage at the max batch size.
        """

        if self.expected_idle_vram_usage is None:
            return float("nan")

        return self.expected_idle_vram_usage

    @property
    def failed_to_profile(self) -> bool:
        """
        :return: Whether the corresponding task is completely invalid since it could not be profiled.
        """

        return math.isnan(self.vram_usage)

    def model_dump(self, **kwargs) -> dict[str, Any]:
        """
        :param kwargs: Standard Pydantic model dump kwargs.
        :return: Dump result of the superclass' method.
        """

        super_dump = super().model_dump(**kwargs)

        if self.vram_usages is not None:
            super_dump[VRAM_USAGES_KEY] = {k: list(v) for k, v in self.vram_usages.items()}

        return super_dump


class InnerTaskProfiles(BaseModel):

    profiles: dict[str, InnerTaskProfile] = {}
    rejected_tasks: list[str] = []

    @property
    def max_layers(self) -> int:
        """
        :return: Maximum number of inner model layers in any one environment.
        """

        return max(profile.number_of_layers for profile in self.profiles.values())

    @property
    def min_layers(self) -> int:
        """
        :return: Minimum number of inner model layers in any one environment.
        """

        return min(profile.number_of_layers for profile in self.profiles.values())

    def max_total_observation_size(self, policy_id: str) -> int:
        """
        :return: The summed observation size of all agents for this policy (excluding the agent mask) with the task that has the most layers.
        """

        if not self.profiles:
            raise ValueError(
                "No profiles to calculate max total observation size. Ensure profiles have been "
                "added before executing the training loop"
            )

        largest_task_name = max(self.profiles, key=lambda k: self.profiles[k].number_of_layers)
        largest_task = self.profiles[largest_task_name]

        return sum(
            [
                v - 1
                for k, v in largest_task.observation_space_sizes.items()
                if k.startswith(POLICIES_TO_PREFIXES[policy_id])
            ]
        )

    @staticmethod
    def validate_size(
        policy_mapping_function: Callable[[str, Any, Any], str],
        size_dictionary: dict[str, int],
        expected_sizes: dict[str, int],
        task_name: str,
    ) -> dict[str, int]:
        """
        :param policy_mapping_function: Function which maps agent IDs to policy IDs.
        :param size_dictionary: Dictionary mapping agent IDs to observation/action space sizes.
        :param expected_sizes: Dictionary mapping policy IDs to expected observation/action space sizes.
        :param task_name: Name of the task being validated.
        :return: Updated expected_sizes dictionary.
        """

        for agent_id, size in size_dictionary.items():
            policy_id = policy_mapping_function(agent_id)  # type: ignore

            if policy_id not in expected_sizes:
                expected_sizes[policy_id] = size

            elif expected_sizes[policy_id] != size:
                raise ValueError(
                    f"The observation space size of agent {agent_id} for task {task_name} does not match the "
                    f"expected size {size}."
                )

        return expected_sizes

    def validate_task_profiles(self, policy_mapping_function: Callable[[str, Any, Any], str]) -> None:
        """
        :param policy_mapping_function: Function which maps agent IDs to policy IDs.
        """

        expected_observation_sizes: dict[str, int] = {}
        expected_action_sizes: dict[str, int] = {}

        for task_name, task_profile in self.profiles.items():
            expected_observation_sizes = self.validate_size(
                policy_mapping_function=policy_mapping_function,
                size_dictionary=task_profile.observation_space_sizes,
                expected_sizes=expected_observation_sizes,
                task_name=task_name,
            )
            expected_action_sizes = self.validate_size(
                policy_mapping_function=policy_mapping_function,
                size_dictionary=task_profile.action_space_sizes,
                expected_sizes=expected_action_sizes,
                task_name=task_name,
            )

    def reject_task(self, task: str) -> None:
        """
        :param task: Task being rejected due to excessive VRAM consumption.
        """

        if task not in self.rejected_tasks:
            self.rejected_tasks.append(task)

    def get_valid_profiles(self, force_tasks: str | list[str] | None = None) -> dict[str, InnerTaskProfile]:
        """
        :param force_tasks: Specific task or list of tasks to always include in valid profiles or None.
        :return: Dictionary mapping inner task names to profiles of the tasks with all tasks that failed to be profiled
        or were rejected removed.
        """

        force_tasks = [force_tasks] if isinstance(force_tasks, str) else force_tasks
        force_tasks = [] if force_tasks is None else force_tasks

        return {
            task: profile
            for task, profile in self.profiles.items()
            if task in force_tasks or (not profile.failed_to_profile and task not in self.rejected_tasks)
        }



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from typing import Any

from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.base_classes import (
    LHOPTBaseObserver,
    LHOPTCheckpointBaseObserver,
)
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class TrainingLoss(GlobalObserver):

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return (
            observation_inputs.training_loss / math.log(num_categories)
            if num_categories is not None
            else observation_inputs.training_loss
        )

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class ValidationLoss(GlobalObserver):

    def __init__(self, keep_dimension: bool = False, **kwargs):
        super().__init__(**kwargs)
        """
        :param keep_dimension: Whether not use validation loss but keep the dimension of this observation to enable client to use this observation for validation loss. If True, training loss will be used instead of validation loss, and will not trigger the requires_validation_loss property.
        """
        self.keep_dimension = keep_dimension

    @property
    def requires_validation_loss(self) -> bool:
        """
        :return: Whether the observation requires validation loss to be calculated.
        """

        return not self.keep_dimension

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        loss = observation_inputs.validation_loss if not self.keep_dimension else observation_inputs.training_loss
        return loss / math.log(num_categories) if num_categories is not None else loss

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class BestObservedValidationLoss(GlobalObserver):

    def __init__(self, keep_dimension: bool = False, **kwargs):
        super().__init__(**kwargs)
        """
        :param keep_dimension: Whether not use validation loss but keep the dimension of this observation to enable client to use this observation for validation loss. If True, training loss will be used instead of validation loss, and will not trigger the requires_validation_loss property.
        """
        self.keep_dimension = keep_dimension

    def _get_best_observed_loss(self, observation_inputs: ObservationInputs) -> float | None:
        """
        :return: Validation loss or training loss.
        """

        return (
            observation_inputs.best_observed_validation_loss
            if not self.keep_dimension
            else observation_inputs.best_observed_training_loss
        )

    @property
    def requires_validation_loss(self) -> bool:
        """
        :return: Whether the observation requires validation loss to be calculated.
        """

        return not self.keep_dimension

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the
        StatisticStorageTypes enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping
        module names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        if self._get_best_observed_loss(observation_inputs) is None:
            return 0
        return (
            self._get_best_observed_loss(observation_inputs) / math.log(num_categories)
            if num_categories is not None
            else self._get_best_observed_loss(observation_inputs)
        )

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no
        kwargs are to be needed.
        """

        return {}


class LossRatio(GlobalObserver):

    def __init__(self, keep_dimension: bool = False, **kwargs):
        super().__init__(**kwargs)
        """
        :param keep_dimension: Whether not use validation loss but keep the dimension of this observation
        to enable client to use this observation for validation loss. If True, training loss will be used
        instead of validation loss, and will not trigger the requires_validation_loss property.
        """
        self.keep_dimension = keep_dimension

    @property
    def requires_validation_loss(self) -> bool:
        """
        :return: Whether the observation requires validation loss to be calculated.
        """

        return not self.keep_dimension

    def _get_loss(self, observation_inputs: ObservationInputs) -> float:
        """
        :return: Validation loss or training loss. if keep_dimension is True,
        training loss will be used instead of validation loss.
        """

        return observation_inputs.validation_loss if not self.keep_dimension else observation_inputs.training_loss

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        if keep_dimension is True, training loss will be used instead of validation loss, loss ratio will be 0 or 1.
        """

        if self._get_loss(observation_inputs) == 0:
            return 0

        return observation_inputs.training_loss / self._get_loss(observation_inputs)

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class LHOPTTrainingLoss(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns a cdf_feature for training loss values.

    This observer use the CDF calculation from the paper and applies CDF transformation using the CDF mean and std.
    """

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """
        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: cdf_feature
        """

        training_loss = observation_inputs.training_loss

        cdf_feature = self._compute_cdf_feature(training_loss)

        self._update_time()

        return cdf_feature

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class LHOPTValidationLoss(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns a cdf_feature for validation loss values.

    This observer use the CDF calculation from the paper and applies CDF transformation using the CDF mean and std.
    """

    def __init__(self, keep_dimension: bool = False, **kwargs):
        super().__init__(**kwargs)
        """
        :param keep_dimension: Whether not use validation loss but keep the dimension of this observation
        to enable client to use this observation for validation loss.
        If True, training loss will be used instead of validation loss,
        and will not trigger the requires_validation_loss property.
        """
        self.keep_dimension = keep_dimension

    @property
    def requires_validation_loss(self) -> bool:
        """
        :return: Whether the observation requires validation loss to be calculated.
        """

        return not self.keep_dimension

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """
        return StatisticStorageTypes.FLOAT

    def _get_loss(self, observation_inputs: ObservationInputs) -> float:
        """
        :return: Validation loss or training loss.
        If keep_dimension is True, training loss will be used instead of validation loss.
        """

        return observation_inputs.validation_loss if not self.keep_dimension else observation_inputs.training_loss

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistics models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: cdf_feature
        """

        cdf_feature = self._compute_cdf_feature(self._get_loss(observation_inputs))

        self._update_time()

        return cdf_feature

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class LHOPTLossRatio(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns two-dimensional observations: [tanh, cdf_feature] for loss ratio values.

    This observer computes the logarithm of the ratio between validation_loss and training_loss,
    providing two features:
    1. tanh(log_ratio) - bounded feature using hyperbolic tangent
    2. cdf_feature - CDF transformed feature using CDF mean and std
    """

    def __init__(self, tanh_bound: float = 10.0, keep_dimension: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.tanh_bound = tanh_bound
        """
        :param keep_dimension: Whether not use validation loss but keep the dimension of this observation
        to enable client to use this observation for validation loss.
        If True, training loss will be used instead of validation loss,
        and will not trigger the requires_validation_loss property.
        """
        self.keep_dimension = keep_dimension

    @property
    def requires_validation_loss(self) -> bool:
        """
        :return: Whether the observation requires validation loss to be calculated.
        """

        return not self.keep_dimension

    def _get_loss(self, observation_inputs: ObservationInputs) -> float:
        """
        :return: Validation loss or training loss.
        If keep_dimension is True, training loss will be used instead of validation loss.
        """

        return observation_inputs.validation_loss if not self.keep_dimension else observation_inputs.training_loss

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 2  # [tanh, cdf_feature]

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> list[int | float]:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistics models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: List of three features: [tanh, cdf_feature]
        """

        log_ratio = self._compute_log_ratio(
            numerator=observation_inputs.training_loss, denominator=self._get_loss(observation_inputs)
        )

        tanh_feature = math.tanh(max(-self.tanh_bound, min(self.tanh_bound, log_ratio)))

        cdf_feature = self._compute_cdf_feature(log_ratio)

        self._update_time()

        return [tanh_feature, cdf_feature]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class PercentileOfLossAtEachCheckpoint(LHOPTCheckpointBaseObserver):
    """
    Observer that computes percentile of loss values at each checkpoint.
    """

    def __init__(
        self,
        checkpoint_interval: int = 100,
        percentile: float = 0.6,
        **kwargs,
    ) -> None:
        """
        :param checkpoint_interval: How often to create checkpoints (in training steps).
        :param percentile: Percentile to compute (0.0 to 1.0).
        :param kwargs: Miscellaneous keyword arguments.
        """
        super().__init__(checkpoint_interval=checkpoint_interval, **kwargs)
        self.percentile = max(0.0, min(1.0, percentile))

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Percentile value of loss at checkpoint.
        """
        training_loss = observation_inputs.training_loss

        # Handle cold start
        self._cold_start(training_loss)

        # Update history
        self._update_history(training_loss)

        # Check if we should create a checkpoint
        if self._should_create_checkpoint():
            # Compute percentile
            sorted_history = sorted(self._history)
            index = int(self.percentile * (len(sorted_history) - 1))
            percentile_value = sorted_history[index]

            self._cached_observation = percentile_value
            return percentile_value
        else:
            # Return last value during warm-up
            self._cached_observation = self.last_value
            return self.last_value



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import enum
from collections import defaultdict

# ======================================================================================================================
#
# SEVERITIES
#
# ======================================================================================================================


class ErrorSeverities(enum.Enum):

    CRITICAL = "critical"
    HIGH = "high"
    WARNING = "warning"
    INFO = "info"

    @classmethod
    def should_mention(cls, severity: enum.Enum) -> bool:
        """
        :param severity: Severity of the error being sent to slack.
        :return: Whether @channel should be mentioned in the message.
        """

        return severity is cls.CRITICAL or severity is cls.HIGH


SEVERITY_LEVELS = defaultdict(
    lambda: 4,
    {ErrorSeverities.CRITICAL: 0, ErrorSeverities.HIGH: 1, ErrorSeverities.WARNING: 2, ErrorSeverities.INFO: 3},
)
SEVERITY_EMOJIS = defaultdict(
    lambda: ":question:",
    {
        ErrorSeverities.CRITICAL: ":rotating_light:",
        ErrorSeverities.HIGH: ":exclamation:",
        ErrorSeverities.WARNING: ":warning:",
        ErrorSeverities.INFO: ":information_source:",
    },
)



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import torch.nn as nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from transformers.activations import ACT2CLS

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class NoOpLRScheduler(LRScheduler):
    """
    A learning rate scheduler that does nothing - preserves the current learning rates.
    This effectively disables any learning rate scheduling while preserving any
    modifications made to the learning rates by agents or other means.
    """

    def __init__(self, optimizer: Optimizer, last_epoch: int = -1) -> None:
        """
        :param optimizer: Optimizer which handles gradient descent.
        :param last_epoch: Epoch at which this scheduler was initialised.
        """
        super().__init__(optimizer=optimizer, last_epoch=last_epoch)

    def get_lr(self) -> list[float]:  # type: ignore
        """
        :return: The current learning rates of the parameter groups exactly as they are.
        This preserves any modifications made by agents or other means.
        """
        return [group["lr"] for group in self.optimizer.param_groups]


# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

PYTORCH_ACTIVATIONS = [
    nn.ReLU.__name__,
    nn.LeakyReLU.__name__,
    nn.PReLU.__name__,
    nn.ReLU6.__name__,
    nn.SELU.__name__,
    nn.CELU.__name__,
    nn.GELU.__name__,
    nn.Sigmoid.__name__,
    nn.Tanh.__name__,
    nn.Hardtanh.__name__,
    nn.Softplus.__name__,
    nn.Softshrink.__name__,
    nn.Softsign.__name__,
    nn.ELU.__name__,
    nn.LogSigmoid.__name__,
    nn.Hardsigmoid.__name__,
    nn.Hardswish.__name__,
    nn.Hardshrink.__name__,
    nn.SiLU.__name__,
    nn.Mish.__name__,
    nn.Softmin.__name__,
    nn.Softmax.__name__,
    nn.LogSoftmax.__name__,
    nn.Tanhshrink.__name__,
]
HUGGINGFACE_ACTIVATIONS = [
    activation.__name__ if not isinstance(activation, tuple) else activation[0].__name__  # type: ignore
    for activation in ACT2CLS.values()
]
ACTIVATIONS = PYTORCH_ACTIVATIONS + HUGGINGFACE_ACTIVATIONS



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import copy
import inspect
from collections import defaultdict
from typing import Any

import torch
import torch.nn as nn
import torch.optim as optim
from accelerate.optimizer import AcceleratedOptimizer
from loguru import logger

from libinephany.utils.constants import PARAMS, SCHEDULER_GROUP_NAME

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

ADAM_OPTIMISERS = [optim.Adam, optim.AdamW]

PARAM_GROUP_LR = "lr"
MODULE_TYPE = "module_type"

LEFTOVER_PARAMS = "leftover_params"
LEFTOVER_MODULE_TYPE = "fixed"
EXP_AVERAGE = "exp_avg"
EXP_AVERAGE_SQ = "exp_avg_sq"
MAX_EXP_AVERAGE_SQ = "max_exp_avg_sq"
BETAS = "betas"
STEP = "step"
EPS = "eps"
LR = "lr"
WEIGHT_DECAY = "weight_decay"
AMSGRAD = "amsgrad"

# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def check_if_tensor_is_xpu(tensor: torch.Tensor | None) -> bool:
    """
    :param tensor: Tensor to check whether it is an XPU tensor.
    :return: Whether the tensor is an XPU tensor.
    """

    try:
        xpu_available = torch.xpu.is_available()
        return xpu_available and tensor.is_xpu

    except AttributeError:
        return False


def tensor_on_local_rank(tensor: torch.Tensor | None) -> bool:
    """
    :param tensor: Tensor to check whether it is owned by the local rank, partially or entirely.
    :return: Whether the tensor is owned by the local rank.
    """

    valid_tensor = tensor is not None and tensor.numel() > 0
    device_index = tensor.device.index if valid_tensor else None

    is_xpu = check_if_tensor_is_xpu(tensor)
    cuda_available = torch.cuda.is_available()

    if valid_tensor and is_xpu:
        current_device = torch.xpu.current_device()
        return device_index == current_device

    elif valid_tensor and cuda_available and tensor.is_cuda:
        current_device = torch.cuda.current_device()
        return device_index == current_device

    return valid_tensor


def check_optimiser_is_adam_variant(optimizer: optim.Optimizer, raise_error: bool = True) -> bool:
    """
    :param optimizer: Optimizer of the inner model to check that it is an Adam variant.
    :param raise_error: Whether an error should be raised if the optimizer is an invalid type or not.
    :return: Whether the optimizer is a valid type.
    """

    if isinstance(optimizer, AcceleratedOptimizer):
        optimizer_unwrapped = optimizer.optimizer
    else:
        optimizer_unwrapped = optimizer

    if not any(isinstance(optimizer_unwrapped, adam_variant) for adam_variant in ADAM_OPTIMISERS):

        if raise_error:
            raise TypeError(
                f"Inner model optimizer is not an Adam variant. Got: {type(optimizer_unwrapped).__name__}. Expected one "
                f"of: {[adam_variant.__name__ for adam_variant in ADAM_OPTIMISERS]}"
            )

        return False

    return True


def compute_adam_optimizer_update_stats(
    optimizer: optim.Optimizer, parameter_group: dict[str, Any], parameters: list[torch.Tensor]
) -> torch.Tensor | None:
    """
    :param optimizer: Adam optimizer to compute update stats for.
    :param parameter_group: Parameter group containing the learning rate and momentum.
    :param parameters: List of parameters to compute update stats from.
    :return: Tensor of the Adam parameter updates or None.
    """
    local_rank_parameters = [p for p in parameters if tensor_on_local_rank(p)]
    first_moment_list = [optimizer.state[p][EXP_AVERAGE].view(-1) for p in local_rank_parameters]
    beta_one_t_list = [torch.pow(parameter_group[BETAS][0], optimizer.state[p][STEP]) for p in local_rank_parameters]
    second_moment_list = [
        optimizer.state[p][MAX_EXP_AVERAGE_SQ if parameter_group[AMSGRAD] else EXP_AVERAGE_SQ].view(-1)
        for p in local_rank_parameters
    ]
    beta_two_t_list = [torch.pow(parameter_group[BETAS][1], optimizer.state[p][STEP]) for p in local_rank_parameters]
    scaled_first_moment_list = [
        first_moment / (1 - beta_one_t) for first_moment, beta_one_t in zip(first_moment_list, beta_one_t_list)
    ]
    scaled_second_moment_list = [
        second_moment / (1 - beta_two_t) for second_moment, beta_two_t in zip(second_moment_list, beta_two_t_list)
    ]
    updates_list = [
        parameter_group[LR] * scaled_first_moment / (torch.sqrt(scaled_second_moment) + parameter_group[EPS])
        for scaled_first_moment, scaled_second_moment in zip(scaled_first_moment_list, scaled_second_moment_list)
    ]
    if isinstance(optimizer, optim.AdamW):
        updates_list = [
            update + parameter_group[LR] * parameter_group[WEIGHT_DECAY] * p.view(-1)
            for update, p in zip(updates_list, local_rank_parameters)
        ]
    return torch.cat(updates_list) if updates_list else None


def filter_optimizer_kwargs(optimizer_class: type[optim.Optimizer], optimizer_kwargs: dict[str, Any]) -> dict:
    """
    :param optimizer_class: Class of optimizer to filter the given kwargs for.
    :param optimizer_kwargs: Optimizer-specific keyword arguments.
    :return: A filtered dictionary containing only valid keyword arguments for the given optimizer class.
    """

    optimizer_signature = inspect.signature(optimizer_class.__init__)

    valid_params = set(optimizer_signature.parameters.keys())

    return {arg_name: arg_value for arg_name, arg_value in optimizer_kwargs.items() if arg_name in valid_params}


def compile_parameter_groups(
    module_to_params: dict[str, list[torch.Tensor]],
    initial_lr: float,
    agent_controlled_modules: dict[str, str],
) -> list[dict[str, Any]]:
    """
    :param module_to_params: Mapping of module names to lists of parameters under that module.
    :param initial_lr: Initial learning rate for the parameter group.
    :param agent_controlled_modules: Dictionary mapping agent modules to their type.
    :return: List of compiled parameter groups.
    """

    return [
        {
            PARAM_GROUP_LR: initial_lr,
            SCHEDULER_GROUP_NAME: module_name,
            PARAMS: params,
            MODULE_TYPE: agent_controlled_modules[module_name],
        }
        for module_name, params in module_to_params.items()
    ]


def create_parameter_groups(
    model: torch.nn.Module,
    agent_controlled_modules: dict[str, str],
    initial_lr: float = 5e-4,
) -> list[dict[str, Any]]:
    """
    :param model: The PyTorch model for which to create parameter groups.
    :param agent_controlled_modules: Dictionary mapping agent modules to their type.
    :param initial_lr: The initial learning rate for all parameter groups.
    :return: A list of dictionaries, each representing a parameter group with 'params',
             'lr', and 'module_name' keys.
    """

    agent_controlled_modules = copy.deepcopy(agent_controlled_modules)

    leftover_params = {}
    module_to_params = defaultdict(list)

    module_splits = [mod.split(".") for mod in agent_controlled_modules.keys()]

    for name, param in model.named_parameters():
        name_split = name.split(".")
        found_module = False

        for module_split in module_splits:
            if name_split[: len(module_split)] == module_split:
                found_module = True

                module_to_params[".".join(module_split)].append(param)
                break

        if not found_module:
            leftover_params[name] = param

    if leftover_params:
        leftover_param_names = ", ".join(list(leftover_params.keys()))
        module_to_params[LEFTOVER_PARAMS] = list(leftover_params.values())
        agent_controlled_modules[LEFTOVER_PARAMS] = LEFTOVER_MODULE_TYPE

        logger.warning(f"The following parameters were not assigned to any group: {leftover_param_names}")

    # type: ignore
    return compile_parameter_groups(
        module_to_params=module_to_params,  # type: ignore
        initial_lr=initial_lr,
        agent_controlled_modules=agent_controlled_modules,
    )


def build_optimizer(
    model: nn.Module,
    agent_controlled_modules: dict[str, str],
    inner_model_optimizer: type[optim.Optimizer],
    initial_learning_rate: float,
    initial_weight_decay: float | None,
    optimizer_kwargs: dict[str, Any],
) -> optim.Optimizer:
    """
    :param model: Model to build an optimizer for.
    :param agent_controlled_modules: Dictionary mapping agent modules to their type.
    :param inner_model_optimizer: Type of the optimizer class to use.
    :param initial_learning_rate: Initial learning rate which all parameter group's learning rates will be set to.
    :param initial_weight_decay: Initial weight decay which all parameter group's weight decays will be set to.
    :param optimizer_kwargs: Optimizer-specific keyword arguments.
    :return: Constructed optimizer.
    """

    initial_weight_decay = initial_weight_decay if initial_weight_decay is not None else 0.0

    param_groups = create_parameter_groups(
        model=model,
        agent_controlled_modules=agent_controlled_modules,
        initial_lr=initial_learning_rate,
    )
    optimizer_kwargs = filter_optimizer_kwargs(optimizer_class=inner_model_optimizer, optimizer_kwargs=optimizer_kwargs)

    optimizer = inner_model_optimizer(
        param_groups, lr=initial_learning_rate, weight_decay=initial_weight_decay, **optimizer_kwargs
    )  # type: ignore

    assert isinstance(optimizer, optim.Optimizer)

    return optimizer



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import os
from typing import Any

import torch
import torch.distributed as dist

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

CUDA = "cuda"
CPU = "cpu"
CUDA_PREFIX = f"{CUDA}:"
MASTER_SCHEDULER_RANK = 0
LOCAL_RANK = "LOCAL_RANK"

# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def is_distributed() -> bool:
    """
    :return: Whether the client is training in a distributed system.
    """

    return dist.is_available() and dist.is_initialized() and dist.get_world_size() > 1


def get_world_size() -> int:
    """
    :return: 1 if distributed training is not being used or the number of ranks in the world if training is distributed.
    """

    if is_distributed():
        return dist.get_world_size()

    else:
        return 1


def get_local_rank() -> int:
    """
    :return: Distributed computing rank of this process.
    """

    if not is_distributed():
        return MASTER_SCHEDULER_RANK

    return dist.get_rank()


def is_scheduler_master_rank() -> bool:
    """
    :return: Whether the current process holds the rank that should control hyperparameter scheduling.
    """

    return get_local_rank() == MASTER_SCHEDULER_RANK


def broadcast_data(data: Any) -> Any:
    """
    :param data: Data to broadcast to all distributed ranks to ensure all ranks hold the same value.
    :return: Same data given which was broadcast or the data that was received from a broadcast emitted by the master
    rank.
    """

    if not is_distributed():
        return data

    broadcast_list = [data if is_scheduler_master_rank() else None]
    dist.broadcast_object_list(broadcast_list, src=MASTER_SCHEDULER_RANK)

    return broadcast_list[0]


def barrier() -> None:
    """
    Calls distributed barrier if possible. Barriers ensure all distributed processes finish processing and sync with
    one another before continuing.
    """

    if is_distributed():
        dist.barrier()


def get_local_device() -> torch.device:
    """
    :return: Local device of the current rank.
    """

    if not is_distributed():
        return torch.device(CUDA if torch.cuda.is_available() else CPU)

    local_device_rank = os.environ.get(LOCAL_RANK, MASTER_SCHEDULER_RANK)
    return torch.device(f"{CUDA_PREFIX}{local_device_rank}" if torch.cuda.is_available() else CPU)



# ======================================================================================================================
#
# GLOBAL OBSERVERS PACKAGE
#
# This package contains all global observer classes used for collecting observations
# across the entire training process, not specific to individual agents.
#
# ======================================================================================================================


from .gradient_observers import GlobalFirstOrderGradients, GlobalSecondOrderGradients
from .hyperparameter_observers import InitialHyperparameters, ModelFamilyOneHot, OptimizerTypeOneHot
from .loss_observers import (
    BestObservedValidationLoss,
    LHOPTLossRatio,
    LHOPTTrainingLoss,
    LHOPTValidationLoss,
    LossRatio,
    PercentileOfLossAtEachCheckpoint,
    TrainingLoss,
    ValidationLoss,
)
from .model_observers import (
    GlobalActivations,
    GlobalLAMBTrustRatio,
    GlobalLogRatioOfUpdateAndPreviousParamNormEnvStepObserver,
    GlobalParameters,
    GlobalParameterUpdates,
    NumberOfLayers,
)
from .progress_observers import EpochsCompleted, ProgressAtEachCheckpoint, StagnationObserver, TrainingProgress

__all__ = [
    InitialHyperparameters.__name__,
    OptimizerTypeOneHot.__name__,
    ModelFamilyOneHot.__name__,
    TrainingLoss.__name__,
    ValidationLoss.__name__,
    BestObservedValidationLoss.__name__,
    LossRatio.__name__,
    GlobalFirstOrderGradients.__name__,
    GlobalSecondOrderGradients.__name__,
    GlobalActivations.__name__,
    GlobalParameterUpdates.__name__,
    GlobalParameters.__name__,
    GlobalLAMBTrustRatio.__name__,
    NumberOfLayers.__name__,
    GlobalLogRatioOfUpdateAndPreviousParamNormEnvStepObserver.__name__,
    TrainingProgress.__name__,
    EpochsCompleted.__name__,
    ProgressAtEachCheckpoint.__name__,
    LHOPTTrainingLoss.__name__,
    LHOPTValidationLoss.__name__,
    LHOPTLossRatio.__name__,
    PercentileOfLossAtEachCheckpoint.__name__,
    StagnationObserver.__name__,
]



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from typing import Any

from pydantic import BaseModel

from libinephany.pydantic_models.configs.hyperparameter_configs import HParamConfigs
from libinephany.pydantic_models.configs.observer_config import AgentObserverConfig

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class BackendResponse(BaseModel):

    inephany_model_status: str
    actions_array: list[Any]
    hidden_states: dict[str, list[Any]] | None = None


class AgentScheduleResponse(BaseModel):
    action: float
    hyperparameter_internal_value: float | int
    agent_type: str
    inephany_model_status: str


class ClientScheduleResponse(BaseModel):
    actions: dict[str, AgentScheduleResponse] = {}

    response_time: float | None = None

    def get_value_mapping(self) -> dict[str, float | int]:
        """
        :return: Dictionary mapping agent IDs to their values.
        """

        return {
            agent_id: agent_response.hyperparameter_internal_value for agent_id, agent_response in self.actions.items()
        }

    def set_values_from_mapping(self, values: dict[str, float | int]) -> None:
        """
        :param values: Dictionary mapping agent IDs to their values.
        """

        for agent_id, value in values.items():
            if agent_id in self.actions:
                self.actions[agent_id].hyperparameter_internal_value = value


class ClientPolicySchemaResponse(BaseModel):
    observation_clipping_threshold: float
    invalid_observation_threshold: float
    invalid_observation_replacement_value: float

    agent_observer_configs: dict[str, AgentObserverConfig]
    hyperparameter_configs: HParamConfigs



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from typing import Any

from libinephany.observations import observation_utils, statistic_trackers
from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.base_classes import LHOPTBaseObserver
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class GlobalActivations(GlobalObserver):

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.ActivationStatistics.__name__]

        return observation_utils.average_tensor_statistics(tensor_statistics=list(statistics.values()))  # type: ignore

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {statistic_trackers.ActivationStatistics.__name__: dict(include_statistics=self.include_statistics)}


class GlobalParameterUpdates(GlobalObserver):

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.ParameterUpdateStatistics.__name__]

        return observation_utils.average_tensor_statistics(tensor_statistics=list(statistics.values()))  # type: ignore

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {statistic_trackers.ParameterUpdateStatistics.__name__: dict(include_statistics=self.include_statistics)}


class GlobalParameters(GlobalObserver):

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.ParameterStatistics.__name__]

        return observation_utils.average_tensor_statistics(tensor_statistics=list(statistics.values()))  # type: ignore

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {statistic_trackers.ParameterStatistics.__name__: dict(include_statistics=self.include_statistics)}


class GlobalLAMBTrustRatio(GlobalObserver):

    def __init__(
        self,
        *,
        use_log_transform: bool = True,
        **kwargs,
    ) -> None:
        """
        :param use_log_transform: Whether to transform the LAMB trust ratio by taking ln(1 + R).
        :param kwargs: Other observation keyword arguments.
        """

        super().__init__(**kwargs)

        self.use_log_transform = use_log_transform

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.LAMBTrustRatioStatistics.__name__]

        return sum(statistics.values()) / len(statistics)  # type: ignore

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {statistic_trackers.LAMBTrustRatioStatistics.__name__: dict(use_log_transform=self.use_log_transform)}


class NumberOfLayers(GlobalObserver):

    def __init__(
        self,
        *,
        use_log_transform: bool = True,
        trainable_only: bool = False,
        **kwargs,
    ) -> None:
        """
        :param use_log_transform: Whether to transform the return of the Observer by ln(1 + N).
        :param trainable_only: Whether to only count trainable layers.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(**kwargs)

        self.use_log_transform = use_log_transform
        self.trainable_only = trainable_only

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """
        # All statistics are the same, arbitrarily pick the first one
        count = list(tracked_statistics[statistic_trackers.NumberOfLayers.__name__].values())[0]

        if self.use_log_transform:
            return math.log(1 + count)  # type: ignore

        else:
            return count

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {statistic_trackers.NumberOfLayers.__name__: dict(trainable_only=self.trainable_only)}


class GlobalLogRatioOfUpdateAndPreviousParamNormEnvStepObserver(LHOPTBaseObserver):
    def __init__(self, tanh_bound: float = 10.0, **kwargs):
        """
        This observer is used to compute the log ratio of the update and previous parameter norm for the inner step.
        NB: the sample frequency of the statistics was set to 4 in the OpenAI paper.
        """
        super().__init__(**kwargs)
        self.tanh_bound = tanh_bound
        self._previous_param_norm: float | None = None

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 2  # [tanh_feature, cdf_feature]

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistics models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: List containing [raw_log_ratio, cdf_feature].
        """

        update_statistics = tracked_statistics[statistic_trackers.LHOPTParameterUpdateStatistics.__name__]
        param_statistics = tracked_statistics[statistic_trackers.LHOPTParameterStatistics.__name__]
        update_norm = observation_utils.average_tensor_statistics(
            tensor_statistics=[stats for stats in update_statistics.values() if isinstance(stats, TensorStatistics)]
        ).norm

        current_param_norm = observation_utils.average_tensor_statistics(
            tensor_statistics=[stats for stats in param_statistics.values() if isinstance(stats, TensorStatistics)]
        ).norm

        if self._previous_param_norm is None:
            self._previous_param_norm = current_param_norm
            self._compute_cdf_feature(0.0)  # default value since we can't compute log ratio yet
            self._update_time()
            return [0.0, 0.0]
        log_ratio = self._compute_log_ratio(update_norm, self._previous_param_norm)
        tanh_feature = math.tanh(max(-self.tanh_bound, min(self.tanh_bound, log_ratio)))
        cdf_feature = self._compute_cdf_feature(log_ratio)
        self._update_time()
        self._previous_param_norm = current_param_norm

        return [tanh_feature, cdf_feature]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {
            statistic_trackers.LHOPTParameterUpdateStatistics.__name__: dict(
                include_statistics=self.include_statistics, sample_frequency=1
            ),
            statistic_trackers.LHOPTParameterStatistics.__name__: dict(
                include_statistics=self.include_statistics, sample_frequency=1
            ),
        }

    def reset(self) -> None:
        """
        Reset the observer by clearing the previous parameter norm and time series.
        """

        super().reset()
        self._previous_param_norm = None



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import os

import boto3
from loguru import logger

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

S3_URI_SCHEME = "s3://"
S3_BUCKET_SEPARATOR = "/"
S3_BOTO_CLIENT = "s3"
S3_LIST_OBJECTS_V2 = "list_objects_v2"
S3_CONTENTS_KEY = "Contents"
S3_OBJECT_KEY = "Key"

# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def parse_s3_url(s3_url: str) -> tuple[str, str]:
    """
    :param s3_url: S3 URL to parse and extract the bucket and blob names from.
    :return: Tuple of:
        - Bucket name.
        - Blob name.
    """

    if s3_url.startswith(S3_URI_SCHEME):
        s3_url = s3_url.replace(S3_URI_SCHEME, "")

    parts = s3_url.split(S3_BUCKET_SEPARATOR, 1)
    bucket = parts[0]
    key = parts[1] if len(parts) > 1 else ""

    return bucket, key


def download_s3_file(s3_url: str, local_path: str) -> None:
    """
    :param s3_url: S3 URL to download data from.
    :param local_path: Path to save the contents of the download to.
    """

    bucket, key = parse_s3_url(s3_url)

    s3 = boto3.client(S3_BOTO_CLIENT)

    logger.info(f"Downloading {s3_url} to {local_path}...")
    s3.download_file(bucket, key, local_path)
    logger.success("Done.")


def download_s3_prefix(s3_url: str, local_directory: str) -> None:
    """
    :param s3_url: S3 URL with a prefix to download all objects under (e.g. "s3://bucket/some/prefix/").
    :param local_directory: Local directory to recreate the prefix structure into.
    """

    bucket, prefix = parse_s3_url(s3_url)

    if prefix and not prefix.endswith(S3_BUCKET_SEPARATOR):
        prefix += S3_BUCKET_SEPARATOR

    s3 = boto3.client(S3_BOTO_CLIENT)
    paginator = s3.get_paginator(S3_LIST_OBJECTS_V2)

    logger.info(f"Downloading all objects under s3://{bucket}/{prefix} to {local_directory}...")

    count = 0
    for page in paginator.paginate(Bucket=bucket, Prefix=prefix):
        for obj in page.get(S3_CONTENTS_KEY, []):
            key = obj[S3_OBJECT_KEY]
            relative_path = key[len(prefix) :]

            if not relative_path:
                continue

            local_path = os.path.join(local_directory, relative_path)
            os.makedirs(os.path.dirname(local_path), exist_ok=True)

            s3.download_file(bucket, key, local_path)
            count += 1

    logger.success(f"Downloaded {count} file(s).")
